/*********************************************************************
 * Capítulo 3. Tutorial de desarrollo con OpenFL. TicTacToe.
 * Programación Multimedia y Juegos
 * Autor: David Vallejo Fernández    david.vallejo@openflbook.com
 * Autor: Carlos González Morcillo   carlos.gonzalez@openflbook.com
 * Autor: David Frutos Talavera      david.frutos@openflbook.com
 *
 * You can redistribute and/or modify this file under the terms of the
 * GNU General Public License ad published by the Free Software
 * Foundation, either version 3 of the License, or (at your option)
 * and later version. See <http://www.gnu.org/licenses/>.
 *
 * This file is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  
 *********************************************************************/

package book.openfl.tictactoe.ia;

import Sys;

// Enumeración para las opciones disponibles en una casilla.
// Empty ------> Casilla vacía.
// Player_O ---> Casilla con O.
// Player_X ---> Casilla con X.
enum TPlayer { Empty; Player_O; Player_X; }

class Board {

  // Estado del tablero.
  // - X O
  // X X O
  // - O -
  private var _pieces:Array<TPlayer>;

  public function new () {
    _pieces = new Array<TPlayer>();
    for (i in 0...9)
      _pieces[i] = Empty;
  }

  public function getPieces () : Array<TPlayer> {
    return _pieces;
  }

  // No comprueba si el movimiento es válido.
  // move representa la casilla [0..8].
  // player representa al jugador.
  public function makeMove (move:Int, player:TPlayer) : Void {
    _pieces[move] = player;
  }

  // Deshace un movimiento previamente realizado.
  public function undoMove (move:Int) : Void {
    _pieces[move] = Empty;
  }

  // Devuelve, si existe, el ganador de una partida
  // considerando el estado actual del tablero.
  // Si no hay ganador, devuelve Empty.
  public function getWinner () : TPlayer {
    var winningStates:Array<Array<Int>> = 
      [
       [0, 1, 2], [3, 4, 5], [6, 7, 8],
       [0, 3, 6], [1, 4, 7], [2, 5, 8],
       [0, 4, 8], [2, 4, 6],
       ];

    for (st in winningStates) {
      var aux:Array<TPlayer> = [_pieces[st[0]], 
				_pieces[st[1]], 
				_pieces[st[2]]];
      if (_pieces[st[0]] != Empty && allEqual(aux))
	return _pieces[st[0]];
    }
    return Empty;
  }

  // Devuelve una lista que contiene las casillas
  // con los movimientos posibles.
  public function getValidMoves () : Array<Int> {
    var validMoves:Array<Int> = new Array<Int>();
    for (i in 0...9)
      if (_pieces[i] == Empty)
	validMoves.push(i);

    return validMoves;
  }

  // Comprueba si el juego ha terminado...
  // ¿Existe algún movimiento válido?
  // ¿Hay un ganador?
  public function isGameOver () : Bool {
    return (getValidMoves().length == 0 || getWinner() != Empty);
  }

  // Impresión del estado del tablero.
  public function prettyPrint () : Void {
    var i:Int = 1;
    Sys.println("");
    for (p in _pieces) {
      Sys.print(p + "      \t");
      if (i % 3 == 0)
	Sys.println("\n");
      i++;
    }
  }

  public function copy () : Board {
    var copy:Board = new Board();
    for (i in 0..._pieces.length)
      copy.getPieces()[i] = _pieces[i];
    return copy;
  }

  // Comprueba si todos los elementos de array son iguales.
  public function allEqual (array:Array<TPlayer>) : Bool {
    return ((array[0] == array[1]) && (array[1] == array[2]));
  }

}