/*********************************************************************
 * Capítulo 3. Tutorial de desarrollo con OpenFL. TicTacToe.
 * Programación Multimedia y Juegos
 * Autor: David Vallejo Fernández    david.vallejo@openflbook.com
 * Autor: Carlos González Morcillo   carlos.gonzalez@openflbook.com
 * Autor: David Frutos Talavera      david.frutos@openflbook.com
 *
 * You can redistribute and/or modify this file under the terms of the
 * GNU General Public License ad published by the Free Software
 * Foundation, either version 3 of the License, or (at your option)
 * and later version. See <http://www.gnu.org/licenses/>.
 *
 * This file is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.  
 *********************************************************************/

package book.openfl.tictactoe.ia;

import Sys;
import book.openfl.tictactoe.ia.Board;
// Variante Negamax del algoritmo MiniMax.
// Tic-tac-toe es un zero-sum game.
class MiniMax {

  static inline var INFINITY = 99999;
  private var _bestMove:Int;          // Mejor movimiento.
  private var _maxDepth:Int;          // Profundidad de búsqueda.

  public function new (maxDepth:Int) {
    _bestMove = -1;
    _maxDepth = maxDepth;
  }

  // Construye recursivamente el árbol hasta una profundidad depth.
  // Devuelve la mejor puntuación.
  public function buildTreeRec (board:Board, currentPlayer:TPlayer, 
				depth:Int) :Int {
    // Para en profundidad maxDepth.
    if (depth > _maxDepth) return 0;

    // Para el cambio de turno...
    var otherPlayer:TPlayer;
    if (currentPlayer == Player_O)
      otherPlayer = Player_X;
    else
      otherPlayer = Player_O;

    var winner:TPlayer = board.getWinner();
    if (winner == currentPlayer) return INFINITY;
    if (winner == otherPlayer)   return -INFINITY;
    if (board.isGameOver())      return 0;
  
    // Lista de posibles movimientos.
    var movelist:Array<Int> = board.getValidMoves();
    // Mejor puntuación.
    var alpha:Int = -INFINITY;
    // Lista con resultados (paralela a movelist)
    var salist:Array<Int> = new Array<Int>();

    // Obtiene los mejores resultados y asocia el mejor movimiento.
    for (i in movelist) {
      // Copia para evaluar el movimiento i.
      var board_copy:Board = board.copy();
      // Movimiento potencial.
      board_copy.makeMove(i, currentPlayer);
      // Cambio de turno (y de signo).
      var subalpha:Int = -buildTreeRec(board_copy, otherPlayer, depth + 1);
      if (alpha < subalpha)
	alpha = subalpha;

      // Añade resultado después de la evaluación.
      if (depth == 0) salist.push(subalpha);
    }

    // Si llega a profundidad 0 y se han explorado todos los
    // sub-árboles, estudia la lista de mejores movimientos
    // y se elige uno de ellos al azar para jugar.
    if (depth == 0) {
      var candidates:Array<Int> = new Array<Int>();
      for (i in 0...salist.length)
	if (salist[i] == alpha)
	  candidates.push(movelist[i]);
      _bestMove = candidates[Std.random(candidates.length)];
    }

    // Devuelve la mejor puntuación.
    // En _bestMove se almacena la mejor opción.
    return alpha;
  }

  public function buildTree (board:Board, currentPlayer:TPlayer) : Int {
    _bestMove = -1;
    // Llamada a la función recursiva.
    var alpha:Int = buildTreeRec(board, currentPlayer, 0);
    return _bestMove;
  }

}